'''

## A modern Pythonic implementation of Avro Phonetic.

---

MIT License

Copyright (c) 2022-present HitBlast

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

# Import third-party modules.
import re
from typing import Any, Dict, List, Optional, Union

# Import local modules.
from . import config
from .utils import validate

# Constants.
PATTERNS: Any = config.AVRO_DICT['data']['patterns']
NON_RULE_PATTERNS: list = [p for p in PATTERNS if 'rules' not in p]
RULE_PATTERNS: list = [p for p in PATTERNS if 'rules' in p]


# The primary parse function for the library.
def parse(*texts: str, in_ascii: bool = False) -> Union[str, List[str]]:
    '''
    ### Parses input text, matches and replaces using avrodict.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    - `*texts (str)`: The text to parse.
    - `in_ascii (bool = False)`: Whether to output in ASCII or not (Unicode).

    Usage:
    ```python
    import avro

    # Unicode
    parsed = avro.parse('ami banglay gan gai')
    print(parsed)

    # ASCII
    parsed_ascii = avro.parse('ami banglay gan gai', in_ascii=True)
    print(parsed_ascii)
    ```
    '''

    def subparse(text: str) -> str:
        # Sanitize text case to meet phonetic comparison standards.
        fixed_text = validate.fix_string_case(text)

        # Prepare output list.
        output = []

        # Cursor end point.
        cur_end = 0

        # Iterate through input text.
        for cur, i in enumerate(fixed_text):
            try:
                i.encode('utf-8')
            except UnicodeDecodeError:
                uni_pass = False
            else:
                uni_pass = True

            match = {"matched": False}

            if not uni_pass:
                cur_end = cur + 1
                output.append(i)

            elif cur >= cur_end and uni_pass:
                match = match_patterns(fixed_text, cur, rule=False)

                if match['matched']:
                    output.append(match['replaced'])
                    cur_end = cur + len(match['found'])

                else:
                    match = match_patterns(fixed_text, cur, rule=True)

                    if match['matched']:
                        cur_end = cur + len(match['found'])
                        replaced = process_rules(
                            rules=match['rules'], fixed_text=fixed_text, cur=cur, cur_end=cur_end
                        )

                        if replaced is not None:
                            output.append(replaced)

                        else:
                            output.append(match['replaced'])

                if not match['matched']:
                    cur_end = cur + 1
                    output.append(i)

        return ''.join(output)

    output = []
    for text in texts:  # Applies to non-keyword arguments.
        output.append(
            subparse(text) if not in_ascii else str(subparse(text).encode('ascii', errors='backslashreplace'))
        )

    return output[0] if len(output) == 1 else output


def reverse(*texts: str) -> Union[str, List[str]]:
    '''
    ### Reverses input text to Roman script typed in English.

    If a valid replacement is found, then it returns the replaced string.
    If no replacement is found, then it instead returns the input text.

    Parameters:
    - `*texts (str)`: The text to reverse.

    Usage:
    ```python
    import avro

    target = avro.reverse('আমার সোনার বাংলা')
    print(target)
    ```
    `output: amar sonar bangla`
    '''

    def subparse(text: str) -> str:
        # Prepare output list.
        output = []

        # Iterate through input text.
        for cur, i in enumerate(text):
            try:
                i.encode('utf-8')
            except UnicodeDecodeError:
                uni_pass = False
            else:
                uni_pass = True

            match = {"matched": False}

            if not uni_pass:
                output.append(i)

            elif uni_pass:
                match = match_patterns(text, cur, rule=False, reversed=True)

                if match['matched']:
                    if match['reversed']:
                        output.append(match['reversed'])
                    else:
                        output.append(match['found'])

                if not match['matched']:
                    output.append(i)

        return ''.join(output)

    text_segments = []
    output = []

    # Split using regex to remove noise.
    regex_pattern = "(\\s|\\.|,|\\?|\\।|\\-|;|')"
    compiled_regex = re.compile(regex_pattern, re.UNICODE)

    for text in texts:  # Applies to non-keyword arguments.
        exceptions = config.EXCEPTIONS.get(text, None)

        if not exceptions:
            separated_texts = compiled_regex.split(text)

            for separated_text in separated_texts:
                text_segments.append(subparse(separated_text))

            output.append(''.join(text_segments))
            text_segments = []

        else:
            output.append(exceptions)

    return output[0] if len(output) == 1 else output


def match_patterns(
    fixed_text: str, cur: int = 0, rule: bool = False, reversed: bool = False
) -> Dict[str, Any]:
    '''
    ### Matches given text at cursor position with rule / non rule patterns.

    Returns a dictionary of three elements:

    - `matched` - Bool: depending on if match found.
    - `found` - string/None: Value of matched pattern's 'find' key or none.
    - `replaced` - string: Replaced string if match found else input string at cursor.
    '''

    rule_type = NON_RULE_PATTERNS if not rule else RULE_PATTERNS
    pattern = exact_find_in_pattern(fixed_text, reversed, cur, rule_type)

    if len(pattern) > 0:
        if not rule:
            return {
                "matched": True,
                "found": pattern[0].get('find', None),
                "replaced": pattern[0].get('replace', None),
                "reversed": reverse_with_rules(cur, fixed_text, pattern[0].get('reverse', None)),
            }
        else:
            return {
                "matched": True,
                "found": pattern[0]['find'],
                "replaced": pattern[0]['replace'],
                "rules": pattern[0]['rules'],
            }
    else:
        if not rule:
            return {"matched": False, "found": None, "replaced": fixed_text[cur]}
        else:
            return {"matched": False, "found": None, "replaced": fixed_text[cur], "rules": None}


def exact_find_in_pattern(
    fixed_text: str, reversed: bool, cur: int = 0, patterns: Any = PATTERNS
) -> List[Dict[str, Any]]:
    '''
    ### Returns pattern items that match given text, cursor position and pattern.
    '''

    if reversed:
        return [
            x
            for x in patterns
            if (cur + len(x['replace']) <= len(fixed_text))
            and x['replace'] == fixed_text[cur : (cur + len(x['replace']))]
        ]

    return [
        x
        for x in patterns
        if x.get('find', None)
        and (cur + len(x['find']) <= len(fixed_text))
        and x['find'] == fixed_text[cur : (cur + len(x['find']))]
    ]


def reverse_with_rules(cursor: int, fixed_text: str, text_reversed: str) -> str:
    '''
    ### Enhances the word with rules for reverse-parsing.
    '''

    added_suffix = ''

    if not (
        fixed_text[cursor] in config.AVRO_KAR
        or fixed_text[cursor] in config.AVRO_SHORBORNO
        or fixed_text[cursor] in config.AVRO_IGNORE
        or len(fixed_text) == cursor + 1
    ):
        added_suffix = 'o'

    try:
        if fixed_text[cursor + 1] in config.AVRO_KAR:
            added_suffix = ''
        if fixed_text[cursor + 2] in config.AVRO_KAR and not cursor == 0:
            added_suffix = ''

    except IndexError:
        pass

    return text_reversed if not text_reversed else text_reversed + added_suffix


def process_rules(rules: Dict[str, Any], fixed_text: str, cur: int = 0, cur_end: int = 1) -> Optional[str]:
    '''
    ### Process rules matched in pattern and returns suitable replacement.

    If any rule's condition is satisfied, output the rules "replace",
    else output None.
    '''

    replaced = ''

    # Iterate through rules.
    for rule in rules:
        matched = False

        for match in rule['matches']:
            matched = process_match(match, fixed_text, cur, cur_end)

            if not matched:
                break

        if matched:
            replaced = rule['replace']
            break

    return replaced if matched else None


def process_match(match: Any, fixed_text: str, cur: int, cur_end: int) -> bool:
    '''
    ### Processes a single match in rules.
    '''

    # Initial/default value for replace.
    replace = True

    # Set check cursor depending on match['type']
    chk = cur - 1 if match['type'] == 'prefix' else cur_end

    # Set scope based on whether scope is negative.
    if match['scope'].startswith('!'):
        scope = match['scope'][1:]
        negative = True
    else:
        scope = match['scope']
        negative = False

    # Let the matching begin!
    if scope == 'punctuation':
        if (
            not (
                (chk < 0 and match['type'] == 'prefix')
                or (chk >= len(fixed_text) and match['type'] == 'suffix')
                or validate.is_punctuation(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == 'vowel':
        if (
            not (
                (
                    (chk >= 0 and match['type'] == 'prefix')
                    or (chk < len(fixed_text) and match['type'] == 'suffix')
                )
                and validate.is_vowel(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == 'consonant':
        if (
            not (
                (
                    (chk >= 0 and match['type'] == 'prefix')
                    or (chk < len(fixed_text) and match['type'] == 'suffix')
                )
                and validate.is_consonant(fixed_text[chk])
            )
            != negative
        ):
            replace = False

    elif scope == 'exact':
        if match['type'] == 'prefix':
            exact_start = cur - len(match['value'])
            exact_end = cur
        else:
            exact_start = cur_end
            exact_end = cur_end + len(match['value'])

        if not validate.is_exact(match['value'], fixed_text, exact_start, exact_end, negative):
            replace = False

    return replace



'''

## Validation functions (avro.utils) for avro.py

---

MIT License

Copyright (c) 2022-present HitBlast

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


# Import local modules.
from avro import config


# Functions.
def is_vowel(text: str) -> bool:
    '''
    ### Check if given string is a vowel.
    '''

    return text.lower() in config.AVRO_VOWELS


def is_consonant(text: str) -> bool:
    '''
    ### Check if given string is a consonant.
    '''

    return text.lower() in config.AVRO_CONSONANTS


def is_number(text: str) -> bool:
    '''
    ### Check if given string is a number.
    '''

    return text.lower() in config.AVRO_NUMBERS


def is_punctuation(text: str) -> bool:
    '''
    ### Check if given string is a punctuation.
    '''

    return not (text.lower() in config.AVRO_VOWELS or text.lower() in config.AVRO_CONSONANTS)


def is_case_sensitive(text: str) -> bool:
    '''
    ### Check if given string is case sensitive.
    '''

    return text.lower() in config.AVRO_CASESENSITIVES


def is_exact(needle: str, haystack: str, start: int, end: int, matchnot: bool) -> bool:
    '''
    ### Check exact occurrence of needle in haystack.
    '''

    return (start >= 0 and end < len(haystack) and haystack[start:end] == needle) != matchnot


def fix_string_case(text: str) -> str:
    '''
    ### Converts case-insensitive characters to lower case.

    Case-sensitive characters as defined in config.AVRO_CASESENSITIVES
    retain their case, but others are converted to their lowercase
    equivalents. The result is a string with phonetic-compatible case
    which will the parser will understand without confusion.
    '''

    fixed = []
    for i in text:
        if is_case_sensitive(i):
            fixed.append(i)
        else:
            fixed.append(i.lower())
    return ''.join(fixed)



'''

## The Avro Dictionary (Adapted)

---

MIT License

Copyright (c) 2022-present HitBlast

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


# Imports.
from typing import Dict


# The dictionary variable.
AVRO_DICT: Dict[str, str] = {
    "meta": {
        "file_name": "avrodict.py",
        "file_description": "Provides Avro Dictionary in native Python dictionary. Adapted from avrodict.json",
        "package": "avro.py",
        "license": "MIT License",
        "source": "https://github.com/hitblast/avro.py/blob/main/avro/resources/avrodict.py",
        "adapted_by": "HitBlast",
        "updated": "20230421",
        "encoding": "utf-8"
    },
    "data": {
        "patterns": [
            {
                "find": "bhl",
                "replace": "ভ্ল"
            },
            {
                "find": "psh",
                "replace": "পশ"
            },
            {
                "find": "bdh",
                "replace": "ব্ধ",
                "reverse": "bdh"
            },
            {
                "find": "bj",
                "replace": "ব্জ",
                "reverse": "bj"
            },
            {
                "find": "bd",
                "replace": "ব্দ",
                "reverse": "bd"
            },
            {
                "find": "bb",
                "replace": "ব্ব",
                "reverse": "bb"
            },
            {
                "find": "bl",
                "replace": "ব্ল",
                "reverse": "bl"
            },
            {
                "find": "bh",
                "replace": "ভ",
                "reverse": "bh"
            },
            {
                "find": "vl",
                "replace": "ভ্ল",
                "reverse": "vl"
            },
            {
                "find": "b",
                "replace": "ব",
                "reverse": "b"
            },
            {
                "find": "v",
                "replace": "ভ",
                "reverse": "bh"
            },
            {
                "find": "cNG",
                "replace": "চ্ঞ",
                "reverse": "cng"
            },
            {
                "find": "cch",
                "replace": "চ্ছ",
                "reverse": "cch"
            },
            {
                "find": "cc",
                "replace": "চ্চ",
                "reverse": "cc"
            },
            {
                "find": "ch",
                "replace": "ছ"
            },
            {
                "find": "c",
                "replace": "চ",
                "reverse": "ch"
            },
            {
                "find": "dhn",
                "replace": "ধ্ন",
                "reverse": "dhn"
            },
            {
                "find": "dhm",
                "replace": "ধ্ম",
                "reverse": "dhm"
            },
            {
                "find": "dgh",
                "replace": "দ্ঘ",
                "reverse": "dgh"
            },
            {
                "find": "ddh",
                "replace": "দ্ধ",
                "reverse": "ddh"
            },
            {
                "find": "dbh",
                "replace": "দ্ভ",
                "reverse": "dv"
            },
            {
                "find": "dv",
                "replace": "দ্ভ",
                "reverse": "dv"
            },
            {
                "find": "dm",
                "replace": "দ্ম",
                "reverse": "dd"
            },
            {
                "find": "DD",
                "replace": "ড্ড",
                "reverse": "dd"
            },
            {
                "find": "Dh",
                "replace": "ঢ",
                "reverse": "dh"
            },
            {
                "find": "dh",
                "replace": "ধ",
                "reverse": "dh"
            },
            {
                "find": "dg",
                "replace": "দ্গ"
            },
            {
                "find": "dd",
                "replace": "দ্দ",
                "reverse": "dd"
            },
            {
                "find": "D",
                "replace": "ড",
                "reverse": "d"
            },
            {
                "find": "d",
                "replace": "দ",
                "reverse": "d"
            },
            {
                "find": "...",
                "replace": "..."
            },
            {
                "find": ".`",
                "replace": ".",
                "reverse": "."
            },
            {
                "find": "..",
                "replace": "।।"
            },
            {
                "find": ".",
                "replace": "।",
                "reverse": "."
            },
            {
                "find": "ghn",
                "replace": "ঘ্ন",
                "reverse": "ghn"
            },
            {
                "find": "Ghn",
                "replace": "ঘ্ন",
                "reverse": "ghn"
            },
            {
                "find": "gdh",
                "replace": "গ্ধ",
                "reverse": "gdh"
            },
            {
                "find": "Gdh",
                "replace": "গ্ধ",
                "reverse": "gdh"
            },
            {
                "find": "gN",
                "replace": "গ্ণ",
                "reverse": "gn"
            },
            {
                "find": "GN",
                "replace": "গ্ণ",
                "reverse": "gn"
            },
            {
                "find": "gn",
                "replace": "গ্ন",
                "reverse": "gn"
            },
            {
                "find": "Gn",
                "replace": "গ্ন",
                "reverse": "gn"
            },
            {
                "find": "gm",
                "replace": "গ্ম"
            },
            {
                "find": "Gm",
                "replace": "গ্ম",
                "reverse": "gm"
            },
            {
                "find": "gl",
                "replace": "গ্ল",
                "reverse": "gl"
            },
            {
                "find": "Gl",
                "replace": "গ্ল",
                "reverse": "gl"
            },
            {
                "find": "gg",
                "replace": "জ্ঞ",
                "reverse": "gg"
            },
            {
                "find": "GG",
                "replace": "জ্ঞ",
                "reverse": "gg"
            },
            {
                "find": "Gg",
                "replace": "জ্ঞ",
                "reverse": "gg"
            },
            {
                "find": "gG",
                "replace": "জ্ঞ",
                "reverse": "gg"
            },
            {
                "find": "gh",
                "replace": "ঘ",
                "reverse": "gh"
            },
            {
                "find": "Gh",
                "replace": "ঘ",
                "reverse": "gh"
            },
            {
                "find": "g",
                "replace": "গ",
                "reverse": "g"
            },
            {
                "find": "G",
                "replace": "গ",
                "reverse": "g"
            },
            {
                "find": "hN",
                "replace": "হ্ণ",
                "reverse": "nn"
            },
            {
                "find": "hn",
                "replace": "হ্ন",
                "reverse": "nn"
            },
            {
                "find": "hm",
                "replace": "হ্ম",
                "reverse": "mm"
            },
            {
                "find": "hl",
                "replace": "হ্ল"
            },
            {
                "find": "h",
                "replace": "হ",
                "reverse": "h"
            },
            {
                "find": "jjh",
                "replace": "জ্ঝ"
            },
            {
                "find": "jNG",
                "replace": "জ্ঞ",
                "reverse": "gg"
            },
            {
                "find": "jh",
                "replace": "ঝ",
                "reverse": "jh"
            },
            {
                "find": "jj",
                "replace": "জ্জ",
                "reverse": "jj"
            },
            {
                "find": "j",
                "replace": "জ",
                "reverse": "j"
            },
            {
                "find": "J",
                "replace": "জ",
                "reverse": "j"
            },
            {
                "find": "kkhN",
                "replace": "ক্ষ্ণ",
                "reverse": "kkhn"
            },
            {
                "find": "kShN",
                "replace": "ক্ষ্ণ",
                "reverse": "kkhn"
            },
            {
                "find": "kkhm",
                "replace": "ক্ষ্ম"
            },
            {
                "find": "kShm",
                "replace": "ক্ষ্ম",
                "reverse": "kkh"
            },
            {
                "find": "kxN",
                "replace": "ক্ষ্ণ",
                "reverse": "kkh"
            },
            {
                "find": "kxm",
                "replace": "ক্ষ্ম",
                "reverse": "kkh"
            },
            {
                "find": "kkh",
                "replace": "ক্ষ",
                "reverse": "kkh"
            },
            {
                "find": "kSh",
                "replace": "ক্ষ",
                "reverse": "kkh"
            },
            {
                "find": "ksh",
                "replace": "কশ"
            },
            {
                "find": "kx",
                "replace": "ক্ষ",
                "reverse": "kkh"
            },
            {
                "find": "kk",
                "replace": "ক্ক",
                "reverse": "kk"
            },
            {
                "find": "kT",
                "replace": "ক্ট",
                "reverse": "kt"
            },
            {
                "find": "kt",
                "replace": "ক্ত",
                "reverse": "kt"
            },
            {
                "find": "kl",
                "replace": "ক্ল",
                "reverse": "kl"
            },
            {
                "find": "ks",
                "replace": "ক্স",
                "reverse": "ks"
            },
            {
                "find": "kh",
                "replace": "খ",
                "reverse": "kh"
            },
            {
                "find": "k",
                "replace": "ক",
                "reverse": "k"
            },
            {
                "find": "lbh",
                "replace": "ল্ভ"
            },
            {
                "find": "ldh",
                "replace": "ল্ধ"
            },
            {
                "find": "lkh",
                "replace": "লখ"
            },
            {
                "find": "lgh",
                "replace": "লঘ"
            },
            {
                "find": "lph",
                "replace": "লফ"
            },
            {
                "find": "lk",
                "replace": "ল্ক",
                "reverse": "lk"
            },
            {
                "find": "lg",
                "replace": "ল্গ"
            },
            {
                "find": "lT",
                "replace": "ল্ট",
                "reverse": "lt"
            },
            {
                "find": "lD",
                "replace": "ল্ড",
                "reverse": "ld"
            },
            {
                "find": "lp",
                "replace": "ল্প",
                "reverse": "lp"
            },
            {
                "find": "lv",
                "replace": "ল্ভ"
            },
            {
                "find": "lm",
                "replace": "ল্ম",
                "reverse": "lm"
            },
            {
                "find": "ll",
                "replace": "ল্ল",
                "reverse": "ll"
            },
            {
                "find": "lb",
                "replace": "ল্ব",
                "reverse": "lb"
            },
            {
                "find": "l",
                "replace": "ল",
                "reverse": "l"
            },
            {
                "find": "mth",
                "replace": "ম্থ"
            },
            {
                "find": "mph",
                "replace": "ম্ফ",
                "reverse": "mf"
            },
            {
                "find": "mbh",
                "replace": "ম্ভ",
                "reverse": "mv"
            },
            {
                "find": "mpl",
                "replace": "মপ্ল"
            },
            {
                "find": "mn",
                "replace": "ম্ন",
                "reverse": "mn"
            },
            {
                "find": "mp",
                "replace": "ম্প",
                "reverse": "mp"
            },
            {
                "find": "mv",
                "replace": "ম্ভ",
                "reverse": "mv"
            },
            {
                "find": "mm",
                "replace": "ম্ম",
                "reverse": "mm"
            },
            {
                "find": "ml",
                "replace": "ম্ল",
                "reverse": "ml"
            },
            {
                "find": "mb",
                "replace": "ম্ব",
                "reverse": "mb"
            },
            {
                "find": "mf",
                "replace": "ম্ফ",
                "reverse": "mf"
            },
            {
                "find": "m",
                "replace": "ম",
                "reverse": "m"
            },
            {
                "find": "0",
                "replace": "০",
                "reverse": "0"
            },
            {
                "find": "1",
                "replace": "১",
                "reverse": "1"
            },
            {
                "find": "2",
                "replace": "২",
                "reverse": "2"
            },
            {
                "find": "3",
                "replace": "৩",
                "reverse": "3"
            },
            {
                "find": "4",
                "replace": "৪",
                "reverse": "4"
            },
            {
                "find": "5",
                "replace": "৫",
                "reverse": "5"
            },
            {
                "find": "6",
                "replace": "৬",
                "reverse": "6"
            },
            {
                "find": "7",
                "replace": "৭",
                "reverse": "7"
            },
            {
                "find": "8",
                "replace": "৮",
                "reverse": "8"
            },
            {
                "find": "9",
                "replace": "৯",
                "reverse": "9"
            },
            {
                "find": "NgkSh",
                "replace": "ঙ্ক্ষ",
                "reverse": "ngkh"
            },
            {
                "find": "Ngkkh",
                "replace": "ঙ্ক্ষ",
                "reverse": "ngkh"
            },
            {
                "find": "NGch",
                "replace": "ঞ্ছ",
                "reverse": "ngch"
            },
            {
                "find": "Nggh",
                "replace": "ঙ্ঘ"
            },
            {
                "find": "Ngkh",
                "replace": "ঙ্খ",
                "reverse": "ngkh"
            },
            {
                "find": "NGjh",
                "replace": "ঞ্ঝ"
            },
            {
                "find": "ngOU",
                "replace": "ঙ্গৌ"
            },
            {
                "find": "ngOI",
                "replace": "ঙ্গৈ"
            },
            {
                "find": "Ngkx",
                "replace": "ঙ্ক্ষ",
                "reverse": "ngkh"
            },
            {
                "find": "NGc",
                "replace": "ঞ্চ",
                "reverse": "nch"
            },
            {
                "find": "nch",
                "replace": "ঞ্ছ",
                "reverse": "ngch"
            },
            {
                "find": "njh",
                "replace": "ঞ্ঝ"
            },
            {
                "find": "ngh",
                "replace": "ঙ্ঘ"
            },
            {
                "find": "Ngk",
                "replace": "ঙ্ক",
                "reverse": "ngk"
            },
            {
                "find": "Ngx",
                "replace": "ঙ্ষ"
            },
            {
                "find": "Ngg",
                "replace": "ঙ্গ",
                "reverse": "ngg"
            },
            {
                "find": "Ngm",
                "replace": "ঙ্ম"
            },
            {
                "find": "NGj",
                "replace": "ঞ্জ",
                "reverse": "ngj"
            },
            {
                "find": "ndh",
                "replace": "ন্ধ",
                "reverse": "ndh"
            },
            {
                "find": "nTh",
                "replace": "ন্ঠ",
                "reverse": "nth"
            },
            {
                "find": "NTh",
                "replace": "ণ্ঠ",
                "reverse": "nth"
            },
            {
                "find": "nth",
                "replace": "ন্থ",
                "reverse": "nth"
            },
            {
                "find": "nkh",
                "replace": "ঙ্খ",
                "reverse": "ngkh"
            },
            {
                "find": "ngo",
                "replace": "ঙ্গ",
                "reverse": "ngg"
            },
            {
                "find": "nga",
                "replace": "ঙ্গা"
            },
            {
                "find": "ngi",
                "replace": "ঙ্গি"
            },
            {
                "find": "ngI",
                "replace": "ঙ্গী"
            },
            {
                "find": "ngu",
                "replace": "ঙ্গু"
            },
            {
                "find": "ngU",
                "replace": "ঙ্গূ"
            },
            {
                "find": "nge",
                "replace": "ঙ্গে"
            },
            {
                "find": "ngO",
                "replace": "ঙ্গো"
            },
            {
                "find": "NDh",
                "replace": "ণ্ঢ"
            },
            {
                "find": "nsh",
                "replace": "নশ"
            },
            {
                "find": "Ngr",
                "replace": "ঙর"
            },
            {
                "find": "NGr",
                "replace": "ঞর"
            },
            {
                "find": "ngr",
                "replace": "ংর"
            },
            {
                "find": "nj",
                "replace": "ঞ্জ",
                "reverse": "ngj"
            },
            {
                "find": "Ng",
                "replace": "ঙ",
                "reverse": "ng"
            },
            {
                "find": "NG",
                "replace": "ঞ",
                "reverse": "y"
            },
            {
                "find": "nk",
                "replace": "ঙ্ক",
                "reverse": "ngk"
            },
            {
                "find": "ng",
                "replace": "ং",
                "reverse": "ng"
            },
            {
                "find": "nn",
                "replace": "ন্ন",
                "reverse": "nn"
            },
            {
                "find": "NN",
                "replace": "ণ্ণ"
            },
            {
                "find": "Nn",
                "replace": "ণ্ন"
            },
            {
                "find": "nm",
                "replace": "ন্ম",
                "reverse": "nm"
            },
            {
                "find": "Nm",
                "replace": "ণ্ম"
            },
            {
                "find": "nd",
                "replace": "ন্দ",
                "reverse": "nd"
            },
            {
                "find": "nT",
                "replace": "ন্ট",
                "reverse": "nt"
            },
            {
                "find": "NT",
                "replace": "ণ্ট",
                "reverse": "nt"
            },
            {
                "find": "nD",
                "replace": "ন্ড",
                "reverse": "nd"
            },
            {
                "find": "ND",
                "replace": "ণ্ড",
                "reverse": "nd"
            },
            {
                "find": "nt",
                "replace": "ন্ত",
                "reverse": "nt"
            },
            {
                "find": "ns",
                "replace": "ন্স"
            },
            {
                "find": "nc",
                "replace": "ঞ্চ",
                "reverse": "nch"
            },
            {
                "find": "n",
                "replace": "ন",
                "reverse": "n"
            },
            {
                "find": "N",
                "replace": "ণ",
                "reverse": "n"
            },
            {
                "find": "OI`",
                "replace": "ৈ",
                "reverse": "oi"
            },
            {
                "find": "OU`",
                "replace": "ৌ",
                "reverse": "ou"
            },
            {
                "find": "O`",
                "replace": "ো",
                "reverse": "o"
            },
            {
                "find": "OI",
                "replace": "ৈ",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            }
                        ],
                        "replace": "ঐ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "ঐ"
                    }
                ],
                "reverse": "oi"
            },
            {
                "find": "OU",
                "replace": "ৌ",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            }
                        ],
                        "replace": "ঔ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "ঔ"
                    }
                ],
                "reverse": "ou"
            },
            {
                "find": "O",
                "replace": "ো",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            }
                        ],
                        "replace": "ও"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "ও"
                    }
                ],
                "reverse": "o"
            },
            {
                "find": "phl",
                "replace": "ফ্ল",
                "reverse": "fl"
            },
            {
                "find": "pT",
                "replace": "প্ট",
                "reverse": "pt"
            },
            {
                "find": "pt",
                "replace": "প্ত",
                "reverse": "pt"
            },
            {
                "find": "pn",
                "replace": "প্ন",
                "reverse": "pn"
            },
            {
                "find": "pp",
                "replace": "প্প",
                "reverse": "pp"
            },
            {
                "find": "pl",
                "replace": "প্ল",
                "reverse": "pl"
            },
            {
                "find": "ps",
                "replace": "প্স",
                "reverse": "ps"
            },
            {
                "find": "ph",
                "replace": "ফ",
                "reverse": "ph"
            },
            {
                "find": "fl",
                "replace": "ফ্ল",
                "reverse": "fl"
            },
            {
                "find": "f",
                "replace": "ফ",
                "reverse": "ph"
            },
            {
                "find": "p",
                "replace": "প",
                "reverse": "p"
            },
            {
                "find": "rri`",
                "replace": "ৃ",
                "reverse": "ri"
            },
            {
                "find": "rri",
                "replace": "ৃ",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            }
                        ],
                        "replace": "ঋ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "ঋ"
                    }
                ],
                "reverse": "ri"
            },
            {
                "find": "rrZ",
                "replace": "রর‍্য"
            },
            {
                "find": "rry",
                "replace": "রর‍্য"
            },
            {
                "find": "rZ",
                "replace": "র‍্য",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x"
                            }
                        ],
                        "replace": "্র্য"
                    }
                ]
            },
            {
                "find": "ry",
                "replace": "র‍্য",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x"
                            }
                        ],
                        "replace": "্র্য"
                    }
                ]
            },
            {
                "find": "rr",
                "replace": "রর",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!vowel"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "r"
                            },
                            {
                                "type": "suffix",
                                "scope": "!punctuation"
                            }
                        ],
                        "replace": "র্"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r"
                            }
                        ],
                        "replace": "্রর"
                    }
                ]
            },
            {
                "find": "Rg",
                "replace": "ড়্গ"
            },
            {
                "find": "Rh",
                "replace": "ঢ়"
            },
            {
                "find": "R",
                "replace": "ড়",
                "reverse": "r"
            },
            {
                "find": "r",
                "replace": "র",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "r"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "y"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "w"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "x"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "Z"
                            }
                        ],
                        "replace": "্র"
                    }
                ]
            },
            {
                "find": "shch",
                "replace": "শ্ছ",
                "reverse": "sch"
            },
            {
                "find": "ShTh",
                "replace": "ষ্ঠ",
                "reverse": "sth"
            },
            {
                "find": "Shph",
                "replace": "ষ্ফ",
                "reverse": "sf"
            },
            {
                "find": "Sch",
                "replace": "শ্ছ",
                "reverse": "sch"
            },
            {
                "find": "skl",
                "replace": "স্ক্ল"
            },
            {
                "find": "skh",
                "replace": "স্খ"
            },
            {
                "find": "sth",
                "replace": "স্থ",
                "reverse": "sth"
            },
            {
                "find": "sph",
                "replace": "স্ফ",
                "reverse": "sf"
            },
            {
                "find": "shc",
                "replace": "শ্চ",
                "reverse": "scch"
            },
            {
                "find": "sht",
                "replace": "শ্ত"
            },
            {
                "find": "shn",
                "replace": "শ্ন",
                "reverse": "sn"
            },
            {
                "find": "shm",
                "replace": "শ্ম",
                "reverse": "ss"
            },
            {
                "find": "shl",
                "replace": "শ্ল",
                "reverse": "sl"
            },
            {
                "find": "Shk",
                "replace": "ষ্ক",
                "reverse": "sk"
            },
            {
                "find": "ShT",
                "replace": "ষ্ট",
                "reverse": "st"
            },
            {
                "find": "ShN",
                "replace": "ষ্ণ",
                "reverse": "sn"
            },
            {
                "find": "Shp",
                "replace": "ষ্প",
                "reverse": "sp"
            },
            {
                "find": "Shf",
                "replace": "ষ্ফ",
                "reverse": "sf"
            },
            {
                "find": "Shm",
                "replace": "ষ্ম",
                "reverse": "sm"
            },
            {
                "find": "spl",
                "replace": "স্প্ল"
            },
            {
                "find": "sk",
                "replace": "স্ক",
                "reverse": "sk"
            },
            {
                "find": "Sc",
                "replace": "শ্চ",
                "reverse": "scch"
            },
            {
                "find": "sT",
                "replace": "স্ট",
                "reverse": "st"
            },
            {
                "find": "st",
                "replace": "স্ত",
                "reverse": "st"
            },
            {
                "find": "sn",
                "replace": "স্ন",
                "reverse": "sn"
            },
            {
                "find": "sp",
                "replace": "স্প",
                "reverse": "sp"
            },
            {
                "find": "sf",
                "replace": "স্ফ",
                "reverse": "sf"
            },
            {
                "find": "sm",
                "replace": "স্ম",
                "reverse": "sh"
            },
            {
                "find": "sl",
                "replace": "স্ল",
                "reverse": "sl"
            },
            {
                "find": "sh",
                "replace": "শ",
                "reverse": "sh"
            },
            {
                "find": "Sc",
                "replace": "শ্চ",
                "reverse": "scch"
            },
            {
                "find": "St",
                "replace": "শ্ত"
            },
            {
                "find": "Sn",
                "replace": "শ্ন",
                "reverse": "sn"
            },
            {
                "find": "Sm",
                "replace": "শ্ম",
                "reverse": "ss"
            },
            {
                "find": "Sl",
                "replace": "শ্ল",
                "reverse": "sl"
            },
            {
                "find": "Sh",
                "replace": "ষ",
                "reverse": "sh"
            },
            {
                "find": "s",
                "replace": "স",
                "reverse": "s"
            },
            {
                "find": "S",
                "replace": "শ",
                "reverse": "sh"
            },
            {
                "find": "oo`",
                "replace": "ু",
                "reverse": "u"
            },
            {
                "find": "oo",
                "replace": "ু",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "উ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "উ"
                    }
                ],
                "reverse": "u"
            },
            # {
            #     "find": "o`"
            #     "replace": ""
            # },
            {
                "find": "oZ",
                "replace": "অ্য"
            },
            {
                "find": "o",
                "replace": "",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "vowel"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "o"
                            }
                        ],
                        "replace": "ও"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "vowel"
                            },
                            {
                                "type": "prefix",
                                "scope": "exact",
                                "value": "o"
                            }
                        ],
                        "replace": "অ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "অ"
                    }
                ]
            },
            {
                "find": "tth",
                "replace": "ত্থ"
            },
            {
                "find": "t``",
                "replace": "ৎ",
                "reverse": "t"
            },
            {
                "find": "TT",
                "replace": "ট্ট",
                "reverse": "tt"
            },
            {
                "find": "Tm",
                "replace": "ট্ম"
            },
            {
                "find": "Th",
                "replace": "ঠ",
                "reverse": "th"
            },
            {
                "find": "tn",
                "replace": "ত্ন"
            },
            {
                "find": "tm",
                "replace": "ত্ম",
                "reverse": "tt"
            },
            {
                "find": "th",
                "replace": "থ",
                "reverse": "th"
            },
            {
                "find": "tt",
                "replace": "ত্ত",
                "reverse": "tt"
            },
            {
                "find": "T",
                "replace": "ট",
                "reverse": "t"
            },
            {
                "find": "t",
                "replace": "ত",
                "reverse": "t"
            },
            {
                "find": "aZ",
                "replace": "অ্যা"
            },
            {
                "find": "AZ",
                "replace": "অ্যা"
            },
            {
                "find": "a`",
                "replace": "া",
                "reverse": "a"
            },
            {
                "find": "A`",
                "replace": "া",
                "reverse": "a"
            },
            {
                "replace": "আ",
                "reverse": "a"
            },
            {
                "replace": "র",
                "reverse": "r"
            },
            {
                "find": "a",
                "replace": "া",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "আ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!exact",
                                "value": "a"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "য়া"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "exact",
                                "value": "a"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "আ"
                    }
                ],
                "reverse": "a"
            },
            {
                "find": "i`",
                "replace": "ি",
                "reverse": "i"
            },
            {
                "replace": "ই",
                "reverse": "i"
            },
            {
                "find": "i",
                "replace": "ি",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ই"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ই"
                    }
                ]
            },
            {
                "find": "I`",
                "replace": "ী",
                "reverse": "i"
            },
            {
                "find": "I",
                "replace": "ী",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঈ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঈ"
                    }
                ],
                "reverse": "i"
            },
            {
                "find": "u`",
                "replace": "ু",
                "reverse": "u"
            },
            {
                "replace": "উ",
                "reverse": "u"
            },
            {
                "find": "u",
                "replace": "ু",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "উ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "উ"
                    }
                ],
                "reverse": "u"
            },
            {
                "find": "U`",
                "replace": "ূ",
                "reverse": "u"
            },
            {
                "find": "U",
                "replace": "ূ",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঊ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঊ"
                    }
                ],
                "reverse": "u"
            },
            {
                "find": "ee`",
                "replace": "ী",
                "reverse": "i"
            },
            {
                "find": "ee",
                "replace": "ী",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঈ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "ঈ"
                    }
                ],
                "reverse": "i"
            },
            {
                "find": "e`",
                "replace": "ে",
                "reverse": "e"
            },
            {
                "replace": "এ",
                "reverse": "e"
            },
            {
                "replace": "ও",
                "reverse": "w"
            },
            {
                "find": "e",
                "replace": "ে",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "এ"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "!exact",
                                "value": "`"
                            }
                        ],
                        "replace": "এ"
                    }
                ],
                "reverse": "e"
            },
            {
                "find": "z",
                "replace": "য",
                "reverse": "z"
            },
            {
                "find": "Z",
                "replace": "্য"
            },
            {
                "find": "y",
                "replace": "্য",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "!consonant"
                            },
                            {
                                "type": "prefix",
                                "scope": "!punctuation"
                            }
                        ],
                        "replace": "য়"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "ইয়"
                    }
                ]
            },
            {
                "find": "Y",
                "replace": "য়",
                "reverse": "y"
            },
            {
                "find": "q",
                "replace": "ক",
                "reverse": "k"
            },
            {
                "find": "w",
                "replace": "ও",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            },
                            {
                                "type": "suffix",
                                "scope": "vowel"
                            }
                        ],
                        "replace": "ওয়"
                    },
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "consonant"
                            }
                        ],
                        "replace": "্ব"
                    }
                ],
                "reverse": "o"
            },
            {
                "find": "x",
                "replace": "ক্স",
                "rules": [
                    {
                        "matches": [
                            {
                                "type": "prefix",
                                "scope": "punctuation"
                            }
                        ],
                        "replace": "এক্স"
                    }
                ],
                "reverse": "ks"
            },
            {
                "find": ":`",
                "replace": ":"
            },
            {
                "find": ":",
                "replace": "ঃ"
            },
            {
                "find": "^`",
                "replace": "^"
            },
            {
                "find": "^",
                "replace": "ঁ",
                "reverse": ""
            },
            {
                "find": ",,",
                "replace": "্‌"
            },
            {
                "find": ",",
                "replace": ","
            },
            {
                "find": "$",
                "replace": "৳"
            },
            # {
            #     "find": "`",
            #     "replace": ""
            # }
        ],

        # Remapped words
        "exceptions": {
            "ফেসবুক": "Facebook",
            "গুগল": "Google",
            "উইকিপিডিয়া": "Wikipedia",
            "হোয়াটসঅ্যাপ": "Whatsapp",
            "টুইটার": "Twitter",
            "লিঙ্কডইন": "Linkedin"
        },

        # Constant values
        "vowel": "aeiou",
        "consonant": "bcdfghjklmnpqrstvwxyz",
        "casesensitive": "oiudgjnrstyz",
        "number": "0123456789",

        # For reverse parsing
        "shorborno": "অআইঈউঊএঐওঔ",
        "kar": {
            'া',
            'ি',
            'ী',
            'ু',
            'ূ',
            'ৃ',
            'ে',
            'ৈ',
            'ো',
            'ৌ'
        },

        # Ignored symbols
        "ignore": {
            "ঁ",
            "।",
            "?",
            ".",
            "-",
            ";"
        }
    }
}



'''

## Test cases for avro.main

---

MIT License

Copyright (c) 2022-present HitBlast

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


# Import third-party modules.
import os
import sys

# Add support layer for accessing the primary package.
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir)))

# Import local modules.
import avro
from avro.config import AVRO_DICT


# Test functions for this file.
def test_patterns_without_rules_from_config() -> None:
    '''
    ### Tests all patterns from config that don't have rules.
    '''

    for pattern in AVRO_DICT['data']['patterns']:
        if 'rules' not in pattern:
            if pattern.get('find', None) is not None:
                assert pattern['replace'] == avro.parse(pattern['find'])


def test_patterns_without_rules_not_from_config() -> None:
    '''
    ### Tests all patterns not from config that don't have rules.

    This test is done in addition to
    test_patterns_without_rules_from_config() to ensure that text
    passed manually to avro.parse are properly parsed when they
    don't exact match a pattern that has no rules specified.
    '''

    conjunctions = {
        "ভ্ল": avro.parse("bhl"),
        "ব্ধ": avro.parse("bdh"),
        "ড্ড": avro.parse("DD"),
        "স্তব্ধ বক": avro.parse("stbdh bk"),  # Stunned stork!
    }

    for key, value in conjunctions.items():
        assert key == value


def test_patterns_numbers() -> None:
    '''
    ### Test patterns - numbers
    '''

    numbers = {
        "০": avro.parse("0"),
        "১": avro.parse("1"),
        "২": avro.parse("2"),
        "৩": avro.parse("3"),
        "৪": avro.parse("4"),
        "৫": avro.parse("5"),
        "৬": avro.parse("6"),
        "৭": avro.parse("7"),
        "৮": avro.parse("8"),
        "৯": avro.parse("9"),
        "১১২": avro.parse("112"),
    }

    for key, value in numbers.items():
        assert key == value


def test_patterns_punctuations() -> None:
    '''
    ### Tests patterns - punctuations
    '''

    punctuations = {"।": avro.parse("."), "।।": avro.parse(".."), "...": avro.parse("...")}

    for key, value in punctuations.items():
        assert key == value


def test_patterns_with_rules_svaravarna() -> None:
    '''
    ### Test patterns - with rules - svaravarna / shoroborno (derived from Bengali)
    '''

    svaravarna = {
        "অ": avro.parse("o"),
        "আ": avro.parse("a"),
        "ই": avro.parse("i"),
        "ঈ": avro.parse("I"),
        "উ": avro.parse("u"),
        "ঊ": avro.parse("oo"),
        "ঊ": avro.parse("U"),
        "এ": avro.parse("e"),
        "ঐ": avro.parse("OI"),
        "ও": avro.parse("O"),
        "ঔ": avro.parse("OU"),
    }

    for key, value in svaravarna.items():
        assert key == value


def test_non_ascii() -> None:
    '''
    ### Test parser response for non ascii characters.
    Parser should return any non-ascii characters that is passed to it.
    '''

    non_ascii = {
        'ব': avro.parse('ব'),
        'অভ্র': avro.parse('অভ্র'),
        'বআবা গো': avro.parse('বaba gO'),  # Mixed strings.
        'আমি বাংলায় গান গাই': avro.parse('aমি বাংলায় gaন গাi'),
    }

    for key, value in non_ascii.items():
        assert key == value


def test_words_with_punctuations() -> None:
    '''
    ### Test parsing of words with punctuations.
    '''

    words_with_punctuations = {
        'আয়রে,': avro.parse('ayre,'),
        'ভোলা': avro.parse('bhOla'),
        'খেয়াল': avro.parse('kheyal'),
        'খোলা': avro.parse('khOla'),
    }

    for key, value in words_with_punctuations.items():
        assert key == value


def tests_sentences_with_default() -> None:
    '''
    ### Test parsing of sentences (Unicode).
    '''

    assert 'আমি বাংলায় গান গাই' == avro.parse('ami banglay gan gai')


def test_sentences_with_ascii_flag() -> None:
    '''
    ### Test parsing of sentences (ASCII).
    '''

    assert str(
        b'\\u0986\\u09ae\\u09bf \\u09ac\\u09be\\u0982\\u09b2\\u09be\\u09df'
        + b' \\u0997\\u09be\\u09a8 \\u0997\\u09be\\u0987'
    ) == avro.parse('ami banglay gan gai', in_ascii=True)


def test_reverse_func() -> None:
    '''
    ### Test reverse-parsing with sentences.
    '''

    assert 'ami banglay gan gai' == avro.reverse('আমি বাংলায় গান গাই')
    assert 'rahim, tomake korim dakche. ekhon ki rowna debe?' == avro.reverse(
        'রাহিম, তোমাকে করিম ডাকছে। এখন কি রওনা দেবে?'
    )



'''

## Test cases for avro.utils.validate

---

MIT License

Copyright (c) 2022-present HitBlast

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the 'Software'), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


# Import third-party modules.
import os
import sys

# Add support layer for accessing the primary package.
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir)))

# Import local modules.
from avro.utils import validate


# Set up test environments.
vowels = 'aeiou'
consonants = 'bcdfghjklmnpqrstvwxyz'
numbers = '0123456789'


# Test functions for this file.
def test_is_consonant() -> None:
    '''
    ### Test that consonants are correctly identified.
    '''

    for i in consonants + consonants.upper():
        assert validate.is_consonant(i)

    for i in vowels + numbers:
        assert not validate.is_consonant(i)


def test_is_number() -> None:
    '''
    ### Test that numbers are correctly identified.
    '''

    for i in numbers:
        assert validate.is_number(i)

    for i in vowels + consonants:
        assert not validate.is_number(i)


def test_is_vowel() -> None:
    '''
    ### Test that vowels are correctly identified.
    '''

    for i in vowels + vowels.upper():
        assert validate.is_vowel(i)

    for i in consonants + numbers:
        assert not validate.is_vowel(i)


def test_is_punctuation() -> None:
    '''
    ### Test that punctuations are correctly identified.

    Anything that is neither a number, nor vowel nor consonant is
    identified as a punctuation.
    '''

    for i in '`~!@#$%^&*()-_=+\\|[{}]\'",<.>/?':
        assert validate.is_punctuation(i)
        assert not validate.is_vowel(i)
        assert not validate.is_consonant(i)
        assert not validate.is_number(i)


def test_fix_string_case() -> None:
    '''
    ### Test phonetic-compatible case-transformations of strings.

    This ensures validate.fix_strings function works as
    expected. It should properly change text to lowercase but
    retain case-sensitive characters defined in config as
    uppercase.
    '''

    assert validate.fix_string_case('ABOL taBOl') == 'abOl tabOl'
    assert validate.fix_string_case('KhiCuRi') == 'khicuRi'
    assert validate.fix_string_case('KaTh-BuRO') == 'kaTh-buRO'
    assert validate.fix_string_case('raMgoRurer Chana') == 'ramgoRurer chana'


def test_is_exact() -> None:
    '''
    ### Test exact search response of needle in haystack.
    '''

    assert validate.is_exact('abcd', 'abcdefgh', 0, 4, False)
    assert not validate.is_exact('abcd', 'abcdefgh', 0, 4, True)
    assert not validate.is_exact('bcd', 'abcdefgh', 0, 4, False)
    assert validate.is_exact('bcd', 'abcdefgh', 0, 4, True)

    assert not validate.is_exact('a', 'a', 1, 2, False)
    assert validate.is_exact('a', 'a', 1, 2, True)



# Import both local and third-party setup modules.
import os
import codecs
from setuptools import setup, find_packages

# Import local modules to fetch version number.
from avro import __version__


# Constants.
here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as fh:
    long_description = "\n" + fh.read()


# Call the setup function.
setup(
    name='avro.py',
    version=__version__,
    description='A modern Pythonic implementation of Avro Phonetic.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    author='HitBlast',
    author_email='hitblastlive@gmail.com',
    url='https://github.com/hitblast/avro.py',
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    package_data={'avro': ['*.md']},
    include_package_data=True,
    license='MIT',
    python_requires=">=3.8",
    keywords=[
        'python',
        'avro',
        'avro phonetic',
        'bangla',
        'bangla phonetics',
        'bengali',
        'bengali phonetics',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
